from __future__ import annotations

from abc import abstractmethod
from ipaddress import IPv4Address, IPv6Address, ip_address
from random import randrange
from typing import Any, Literal, TypeAlias

from .. import MultihostHost, MultihostRole, MultihostUtility
from ..conn import ProcessLogLevel

__all__ = ["Firewall", "Firewalld", "WindowsFirewall"]


HostSpec: TypeAlias = str | MultihostHost | MultihostRole
"""Firewall host/hostname specification."""


ProtocolSpec: TypeAlias = Literal["tcp", "udp"]
"""Firewall protocol specification."""


PortSpec: TypeAlias = int | tuple[int, ProtocolSpec]
"""Firewall port specification."""


class Firewall(MultihostUtility[MultihostHost]):
    """
    Configure host firewall.

    All changes are automatically reverted when a test is finished.
    """

    @property
    @abstractmethod
    def inbound(self) -> FirewallInboundRules:
        """
        Configure firewall inbound rules.

        :return: Inbound rules manager.
        :rtype: FirewallInboundRules
        """
        pass

    @property
    @abstractmethod
    def outbound(self) -> FirewallOutboundRules:
        """
        Configure firewall outbound rules.

        :return: Outbound rules manager.
        :rtype: FirewalldOutboundRules
        """
        pass

    def parse_port_spec(self, spec: PortSpec) -> tuple[int, ProtocolSpec]:
        """
        Parse port specification into (port, protocol) tuple.

        :raises TypeError: When invalid type is given.
        :return: (port, protocol) tuple)
        :rtype: tuple[int, Literal["tcp", "udp"]]
        """
        port: int = 0
        protocol: ProtocolSpec = "tcp"

        if isinstance(spec, int):
            port = spec
            protocol = "tcp"
        elif isinstance(spec, tuple):
            if list(map(type, spec)) == [int, str]:
                port = spec[0]
                protocol = spec[1]
            else:
                raise TypeError(f"Unexpected type: tuple{list(map(type, spec))}")
        else:
            raise TypeError(f"Unexpected type: {type(spec)}")

        if protocol not in ["tcp", "udp"]:
            raise ValueError(f"Unexpected protocol: {protocol}")

        return (port, protocol)

    def parse_host_spec(self, host: HostSpec) -> str:
        """
        Parse host specification into hostname.

        :raises ValueError: When invalid type is given.
        :return: Hostname.
        :rtype: str
        """
        if isinstance(host, str):
            return host
        elif isinstance(host, MultihostHost):
            return host.hostname
        elif isinstance(host, MultihostRole):
            return host.host.hostname
        else:
            raise ValueError(f"Unexpected type of host: {type(host)}")


class FirewallInboundRules(object):
    """
    Configure firewall inbound rules.

    All changes are automatically reverted when a test is finished.
    """

    @abstractmethod
    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Accept incoming traffic on given port.

        .. code-block:: python

            firewall.inbound.accept_port(389)  # missing protocol defaults to "tcp"
            firewall.inbound.accept_port((389, "tcp"))
            firewall.inbound.accept_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass

    @abstractmethod
    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Drop incoming traffic on given port.

        .. code-block:: python

            firewall.inbound.drop_port(389)  # missing protocol defaults to "tcp"
            firewall.inbound.drop_port((389, "tcp"))
            firewall.inbound.drop_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass


class FirewallOutboundRules(object):
    """
    Configure firewall outbound rules.

    All changes are automatically reverted when a test is finished.
    """

    @abstractmethod
    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Accept outgoing traffic on given port.

        .. code-block:: python

            firewall.outbound.accept_port(389)  # missing protocol defaults to "tcp"
            firewall.outbound.accept_port((389, "tcp"))
            firewall.outbound.accept_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass

    @abstractmethod
    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Drop outgoing traffic on given port.

        .. code-block:: python

            firewall.outbound.drop_port(389)  # missing protocol defaults to "tcp"
            firewall.outbound.drop_port((389, "tcp"))
            firewall.outbound.drop_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass

    @abstractmethod
    def accept_host(self, host: HostSpec | list[HostSpec]) -> None:
        """
        Accept outgoing traffic to given host.

        The hostname is resolved to IPv4 and IPv6 addresses via DNS and
        an accept rule is created for each available address.

        .. code-block:: python

            firewall.outbound.accept_host("example.com")
            firewall.outbound.accept_host(["example1.com", "example2.com"])
            firewall.outbound.accept_host(multihost_host)
            firewall.outbound.accept_host(multihost_role)

        :param host: Hostname, MultihostHost or MultihostRole object.
        :type host: HostSpec | list[HostSpec]
        """
        pass

    @abstractmethod
    def drop_host(self, host: HostSpec | list[HostSpec]) -> None:
        """
        Drop outgoing traffic to given host.

        The hostname is resolved to IPv4 and IPv6 addresses via DNS and
        an drop rule is created for each available address.

        .. code-block:: python

            firewall.outbound.drop_host("example.com")
            firewall.outbound.drop_host(["example1.com", "example2.com"])
            firewall.outbound.drop_host(multihost_host)
            firewall.outbound.drop_host(multihost_role)

        :param host: Hostname, MultihostHost or MultihostRole object.
        :type host: HostSpec | list[HostSpec]
        """
        pass


class Firewalld(Firewall):
    """
    Configure firewall using firewalld.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)

        self.__inbound: FirewalldInboundRules = FirewalldInboundRules(self)
        self.__outbound: FirewalldOutboundRules = FirewalldOutboundRules(self)
        self._policies: list[str] = []

        self._priority: int = 30000
        """
        Next priority for new firewall rules, when the priority is
        auto-assigned.

        Given the nature of testing, we start with high priority and then
        decrement the value so rules can be easily overwritten (e.g. accept with
        port). E.g. adding a rule "accept port 80" and then "drop port 80" does
        not remove "accept" rule but the "drop" rule takes precedence.
        """

    def setup(self) -> None:
        """
        Set the firewall up.

        :meta private:
        """
        super().setup()
        self.add_policy(f"test-policy-{randrange(99999)}", ingress="HOST", egress="ANY")
        self.host.conn.exec(["firewall-cmd", "--reload"], log_level=ProcessLogLevel.Error)

    def teardown(self) -> None:
        """
        Revert all firewall changes.

        :meta private:
        """
        for policy in list(self._policies):
            self.remove_policy(policy)
        self.host.conn.exec(["firewall-cmd", "--reload"])
        super().teardown()

    @property
    def inbound(self) -> FirewalldInboundRules:
        """
        Configure firewall inbound rules.

        :return: Inbound rules manager.
        :rtype: FirewallInboundRules
        """
        return self.__inbound

    @property
    def outbound(self) -> FirewalldOutboundRules:
        """
        Configure firewall outbound rules.

        :return: Outbound rules manager.
        :rtype: FirewalldOutboundRules
        """
        return self.__outbound

    @property
    def _next_priority(self) -> int:
        """
        Next auto-assigned priority.

        :raises ValueError: _description_
        :return: _description_
        :rtype: int
        """
        priority = self._priority
        if priority < 0:
            raise ValueError("Too many rules defined")

        self._priority -= 1
        return priority

    @property
    def _default_policy(self) -> str:
        """
        Returns the name of the default policy.

        The default policy is the first one to be added.
        This is usually the one created at ``setup()``.

        If no policy was created (quite strange situation), an exception is raised.
        """
        return self._policies[0]

    def add_policy(
        self,
        name: str,
        *,
        ingress: str | None = None,
        egress: str | None = None,
        priority: int | None = None,
        target: Literal["CONTINUE", "ACCEPT", "DROP", "REJECT"] | None = None,
    ) -> int:
        """
        Add a new (permanent) policy.

        Except for the name, all parameters are optional. When the priority is not provided,
        the next priority is assigned. When the other parameters are not provided,
        no value is assigned to the newly created policy and it defaults to ``firewalld``'s
        defaults.

        :param name: The policy name
        :type name: str
        :param ingress: The ingress zone, not assigned if not provided.
        :type ingress: str | None, optional.
        :param egress: The egress zone, not assigned if not provided.
        :type egress: str | None, optional.
        :param priority: Rule priority, defaults to the next priority.
        :type priority: int | None, optional.
        :param target: Rule target, not assigned if not provided.
        :type target: Literal["CONTINUE", "ACCEPT", "DROP", "REJECT"] | None, optional.
        :return: Policy priority.
        :rtype: int
        """
        self.logger.info(f'Firewalld: adding policy "{name}"')

        cmd = ["firewall-cmd", "--permanent", "--new-policy", name]
        self.host.conn.exec(cmd, log_level=ProcessLogLevel.Error)
        self._policies.append(name)
        cmd[2] = "--policy"

        if priority is None:
            priority = self._next_priority
        self.host.conn.exec([*cmd, "--set-priority", str(priority)], log_level=ProcessLogLevel.Error)

        if ingress is not None:
            self.host.conn.exec([*cmd, "--add-ingress-zone", ingress], log_level=ProcessLogLevel.Error)

        if egress is not None:
            self.host.conn.exec([*cmd, "--add-egress-zone", egress], log_level=ProcessLogLevel.Error)

        if target is not None:
            self.host.conn.exec([*cmd, "--set-target", target], log_level=ProcessLogLevel.Error)

        return priority

    def remove_policy(self, name: str) -> None:
        """
        Remove a (permanent) policy.

        :param name: The name of the policy to be removed.
        :type name: str
        """
        self.logger.info(f'Firewalld: removing policy "{name}"')
        self.host.conn.exec(["firewall-cmd", "--permanent", "--delete-policy", name], log_level=ProcessLogLevel.Error)
        self._policies.remove(name)

    def add_rich_rule(self, rule: str, policy: str | None = None, priority: int | None = None) -> int:
        """
        Add rich rule.

        When the policy is specified, the rule will be added to that policy,
        or to the default policy when not specified.

        The parameter "rule" is the part after "rule priority=X". This part is
        added automatically. That is:

        .. code-block:: console

            $ firewall-cmd [--policy {policy}] --add-rich-rule rule priority={priority} {rule}

        :param rule: Firewalld rich rule.
        :type rule: str
        :param policy: The policy to use.
        :type policy: str | None, optional
        :param priority: Rule priority, defaults to None (= auto-assign next
            value)
        :type priority: int | None, optional
        :return: Rule priority, to be used for rule removal.
        :rtype: int
        """
        if priority is None:
            priority = self._next_priority

        if policy is None:
            policy = self._default_policy

        rule = f"rule priority={priority} {rule}"
        self.logger.info(f'Firewalld: adding rich rule "{rule}"')
        self.host.conn.exec(
            ["firewall-cmd", "--policy", policy, "--add-rich-rule", rule], log_level=ProcessLogLevel.Error
        )

        return priority

    def remove_rich_rule(self, priority: int, rule: str, policy: str | None = None) -> None:
        """
        Remove rich rule.

        The parameter "rule" is the part after "rule priority=X". This part is
        added automatically. That is:

        .. code-block:: console

            $ firewall-cmd --remove-rich-rule rule priority="{priority}" {rule}

        When the policy is specified, the rule will be removev from that policy,
        or from the default policy when not specified.

        :param priority: Rule priority
        :type priority: int
        :param rule: Firewalld rich rule.
        :type rule: str
        :param policy: The policy to use.
        :type policy: str | None, optional
        """
        if policy is None:
            policy = self._default_policy

        rule = f"rule priority={priority} {rule}"
        self.logger.info(f'Firewalld: removing rich rule  "{rule}"')
        self.host.conn.exec(
            ["firewall-cmd", "--policy", policy, "--remove-rich-rule", rule], log_level=ProcessLogLevel.Error
        )


class FirewalldInboundRules(FirewallInboundRules):
    """
    Configure firewall inbound rules using firewalld.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, firewall: Firewalld) -> None:
        """
        :param firewall: Firewalld controller.
        :type firewall: Firewalld
        """
        super().__init__()

        self.firewall: Firewalld = firewall

    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="accept")

    def reject_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Reject incoming traffic on given port.

        This blocks the communication and also sends icmp-host-unreachable.

        .. code-block:: python

            firewall.inbound.reject_port(389)  # missing protocol defaults to "tcp"
            firewall.inbound.reject_port((389, "tcp"))
            firewall.inbound.reject_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        self.__add_port(port, action="reject")

    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="drop")

    def __add_port(
        self,
        port: PortSpec | list[PortSpec],
        *,
        action: Literal["accept", "reject", "drop"],
    ) -> None:
        def __add_port(port: int, protocol: str):
            self.firewall.add_rich_rule(f'port port="{port}" protocol="{protocol}" {action}')

        def __add_service(name: str):
            self.firewall.add_rich_rule(f'service name="{name}" {action}')

        items = port if isinstance(port, list) else [port]
        for item in items:
            if isinstance(item, str):
                __add_service(item)
                continue

            port, protocol = self.firewall.parse_port_spec(item)
            __add_port(port, protocol)


class FirewalldOutboundRules(FirewallOutboundRules):
    """
    Configure firewall outbound rules using firewalld.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, firewall: Firewalld) -> None:
        """
        :param firewall: Firewalld controller.
        :type firewall: Firewalld
        """
        super().__init__()

        self.firewall: Firewalld = firewall

    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="accept")

    def reject_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="reject")

    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="drop")

    def accept_host(self, host: HostSpec | list[HostSpec]) -> None:
        self.__add_host(host, action="accept")

    def reject_host(self, host: HostSpec | list[HostSpec]) -> None:
        """
        Reject outgoing traffic to given host.

        This blocks the communication and also sends icmp-host-unreachable.

        The hostname is resolved to IPv4 and IPv6 addresses via DNS and
        an drop rule is created for each available address.

        .. code-block:: python

            firewall.outbound.drop_host("example.com")
            firewall.outbound.drop_host(["example1.com", "example2.com"])
            firewall.outbound.drop_host(multihost_host)
            firewall.outbound.drop_host(multihost_role)

        :param host: Hostname, MultihostHost or MultihostRole object.
        :type host: HostSpec | list[HostSpec]
        """
        self.__add_host(host, action="reject")

    def drop_host(self, host: HostSpec | list[HostSpec]) -> None:
        self.__add_host(host, action="drop")

    def __add_port(
        self,
        port: PortSpec | list[PortSpec],
        *,
        action: Literal["accept", "reject", "drop"],
    ) -> None:
        items = port if isinstance(port, list) else [port]
        for item in items:
            if isinstance(item, str):
                port = int(item)
                protocol = "tcp"
            else:
                port, protocol = self.firewall.parse_port_spec(item)

            self.firewall.add_rich_rule(f"port port={port} protocol={protocol} {action}")

    def __add_host(
        self,
        host: HostSpec | list[HostSpec],
        *,
        action: Literal["accept", "reject", "drop"],
    ) -> None:
        items = host if isinstance(host, list) else [host]
        for item in items:
            hostname = self.firewall.parse_host_spec(item)
            ipv4s = self.__resolve_hostname(hostname, "A")
            ipv6s = self.__resolve_hostname(hostname, "AAAA")

            self.firewall.logger.info(
                f"Firewalld: adding {action} firewall rule for host {hostname}",
                extra={
                    "data": {
                        "Found IPv4 addresses": ipv4s,
                        "Found IPv6 addresses": ipv6s,
                    }
                },
            )

            for ip in ipv4s:
                self.firewall.add_rich_rule(f"family=ipv4 destination address={ip} {action}")

            for ip in ipv6s:
                self.firewall.add_rich_rule(f"family=ipv6 destination address={ip} {action}")

    def __resolve_hostname(self, hostname: str, type: Literal["A", "AAAA"]) -> list[str]:
        addrs = []
        try:
            ip = ip_address(hostname)
            ip_type = IPv4Address if type == "A" else IPv6Address
            if isinstance(ip, ip_type):
                addrs = [hostname]
        except ValueError:
            result = self.firewall.host.conn.exec(
                ["dig", "+short", "-t", type, hostname], log_level=ProcessLogLevel.Error
            )
            addrs = result.stdout_lines

        return addrs


class WindowsFirewall(Firewall):
    """
    Configure Windows Firewall.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)

        self.__inbound: WindowsFirewallInboundRules = WindowsFirewallInboundRules(self)
        """Manage inbound rules."""

        self.__outbound: WindowsFirewallOutboundRules = WindowsFirewallOutboundRules(self)
        """Manage outbound rules."""

        self._changed: bool = False
        """Did we change anything?"""

        self._rules: list[str] = []
        self._backup: str = "C:\\.mh_firewall.bak.wfw"

    def setup(self):
        """
        Create a backup of current firewall configuration.

        :meta private:
        """
        super().setup()
        self.logger.info(f"Windows Firewall: creating backup at '{self._backup}'")
        self.host.conn.run(
            f"Remove-Item {self._backup}; netsh advfirewall export {self._backup}", log_level=ProcessLogLevel.Error
        )

    def teardown(self):
        """
        Revert all firewall changes.

        :meta private:
        """
        self.logger.info(f"Windows Firewall: restoring from '{self._backup}'")
        self.host.conn.run(
            f"netsh advfirewall reset; netsh advfirewall import {self._backup}", log_level=ProcessLogLevel.Error
        )
        super().teardown()

    @property
    def inbound(self) -> WindowsFirewallInboundRules:
        """
        Configure firewall inbound rules.

        :return: Inbound rules manager.
        :rtype: FirewallInboundRules
        """
        return self.__inbound

    @property
    def outbound(self) -> WindowsFirewallOutboundRules:
        """
        Configure firewall outbound rules.

        :return: Outbound rules manager.
        :rtype: FirewalldOutboundRules
        """
        return self.__outbound

    def add_rule(
        self,
        name: str,
        direction: Literal["inbound", "outbound"],
        action: Literal["allow", "block"],
        args: list[Any],
    ) -> str:
        """
        Add firewall rule.

        Final rule name is constructed as ``"mh/{direction}/block/{name}"``.

        :param name: Rule name.
        :type name: str
        :param direction: Direction
        :type direction: Literal["inbound", "outbound"]
        :param action: Action.
        :type action: Literal["allow", "block"]
        :param args: Additional arguments to New-NetFirewallRule command.
        :type args: list[Any]
        :raises ValueError: If invalid action is given.
        :return: Final rule name.
        :rtype: str
        """
        fullname = f"mh/{direction}/{action}/{name}"
        cmd = ["New-NetFirewallRule", "-DisplayName", fullname, "-Direction", direction, "-Action", action, *args]

        match action:
            case "allow":
                opposite = f"mh/{direction}/block/{name}"
            case "block":
                opposite = f"mh/{direction}/allow/{name}"
            case _:
                raise ValueError(f"Unknown action: {action}")

        if opposite in self._rules:
            self.remove_rule(opposite)

        self.logger.info(f'Windows Firewall: adding rule: {" ".join([str(x) for x in cmd])}')
        self.host.conn.exec(cmd, log_level=ProcessLogLevel.Error)
        self._rules.append(fullname)

        return fullname

    def remove_rule(self, name: str) -> None:
        """
        Remove firewall rule.

        :param name: Complete rule name (e.g. ``mh/inbound/block/389``)
        :type name: str
        """
        cmd = ["Remove-NetFirewallRule", "-DisplayName", name]
        self.logger.info(f'Windows Firewall: removing rule: {" ".join([str(x) for x in cmd])}')
        self.host.conn.exec(cmd, log_level=ProcessLogLevel.Error)
        self._rules.remove(name)


class WindowsFirewallInboundRules(FirewallInboundRules):
    """
    Configure firewall inbound rules using Windows Firewall.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, firewall: WindowsFirewall) -> None:
        """
        :param firewall: WindowsFirewall controller.
        :type firewall: WindowsFirewall
        """
        super().__init__()

        self.firewall: WindowsFirewall = firewall

    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="allow")

    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="block")

    def __add_port(
        self,
        port: PortSpec | list[PortSpec],
        *,
        action: Literal["allow", "block"],
    ) -> None:
        items = port if isinstance(port, list) else [port]
        for item in items:
            port, protocol = self.firewall.parse_port_spec(item)
            self.firewall.add_rule(
                f"{port}/{protocol}", "inbound", action, ["-Protocol", protocol, "-LocalPort", port]
            )


class WindowsFirewallOutboundRules(FirewallOutboundRules):
    """
    Configure firewall outbound rules using Windows Firewall.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, firewall: WindowsFirewall) -> None:
        """
        :param firewall: WindowsFirewall controller.
        :type firewall: WindowsFirewall
        """
        super().__init__()

        self.firewall: WindowsFirewall = firewall

    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="allow")

    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="block")

    def accept_host(self, host: HostSpec | list[HostSpec]) -> None:
        self.__add_host(host, action="allow")

    def drop_host(self, host: HostSpec | list[HostSpec]) -> None:
        self.__add_host(host, action="block")

    def __add_port(
        self,
        port: PortSpec | list[PortSpec],
        *,
        action: Literal["allow", "block"],
    ) -> None:
        items = port if isinstance(port, list) else [port]
        for item in items:
            port, protocol = self.firewall.parse_port_spec(item)
            self.firewall.add_rule(
                f"{port}/{protocol}", "outbound", action, ["-Protocol", protocol, "-RemotePort", port]
            )

    def __add_host(
        self,
        host: HostSpec | list[HostSpec],
        *,
        action: Literal["allow", "block"],
    ) -> None:
        items = host if isinstance(host, list) else [host]
        for item in items:
            hostname = self.firewall.parse_host_spec(item)
            ipv4s = self.__resolve_hostname(hostname, "A")
            ipv6s = self.__resolve_hostname(hostname, "AAAA")

            self.firewall.logger.info(
                f"Windows Firewall: adding {action} firewall rule for host {hostname}",
                extra={
                    "data": {
                        "Found IPv4 addresses": ipv4s,
                        "Found IPv6 addresses": ipv6s,
                    }
                },
            )

            self.firewall.add_rule(f"{hostname}", "outbound", action, ["-RemoteAddress", ",".join([*ipv4s, *ipv6s])])

    def __resolve_hostname(self, hostname: str, type: Literal["A", "AAAA"]) -> list[str]:
        result = self.firewall.host.conn.run(
            f"(Resolve-DnsName -Type {type} -Name {hostname}).IpAddress", log_level=ProcessLogLevel.Error
        )
        return result.stdout_lines



"""SELinux utilities."""

from __future__ import annotations

import re

from .. import MultihostHost, MultihostUtility
from ..conn import ProcessLogLevel

__all__ = [
    "SELinuxContext",
    "SELinux",
]


class SELinuxContext(object):
    """
    Result of ``ls -Z``
    """

    def __init__(self, user: str, role: str, type: str, sensitivity: str, categories: str | None) -> None:
        self.user: str = user
        """
        SELinux user.
        """

        self.role: str = role
        """
        SELinux role.
        """

        self.type: str = type
        """
        SELinux type.
        """

        self.sensitivity: str = sensitivity
        """
        SELinux sensitivity level.
        """

        self.categories: str | None = categories
        """
        SELinux categories.
        """

    def __str__(self) -> str:
        return f"({self.user}:{self.role}:{self.type}:{self.sensitivity}:{self.categories})"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other) -> bool:
        if not isinstance(other, SELinuxContext):
            return False

        return (
            self.user == other.user
            and self.role == other.role
            and self.type == other.type
            and self.sensitivity == other.sensitivity
            and self.categories == other.categories
        )

    @classmethod
    def FromOutput(cls, stdout: str) -> SELinuxContext:
        match = re.match(r"([^ ]+)\s+", stdout)

        if not match:
            raise ValueError("Unexpected value: expecting space separated string")

        selinux_context = match.group(1)
        labels = selinux_context.split(":")

        if len(labels) == 5:
            return cls(labels[0], labels[1], labels[2], labels[3], labels[4])
        elif len(labels) == 4:
            return cls(labels[0], labels[1], labels[2], labels[3], None)
        else:
            raise ValueError(f"Unexpected value: got {len(labels)} labels, expecting 4 or 5")


class SELinux(MultihostUtility[MultihostHost]):
    """
    SELinux utilities
    """

    def get_file_context(self, path: str) -> SELinuxContext | None:
        """
        Gets SELinux file context.

        :param path: File path.
        :type path: str
        :return: SELinux file context
        :rtype: SELinuxContext or None
        """
        self.logger.info(f'Getting SELinux context for "{path}"')
        result = self.host.conn.exec(["ls", "-Z", path], log_level=ProcessLogLevel.Error, raise_on_error=False)

        if result.rc != 0:
            return None

        return SELinuxContext.FromOutput(result.stdout)



from __future__ import annotations


class ArtifactsExceptionGroup(ExceptionGroup):
    """
    One or more exception occurred during artifacts collection.
    """

    ...


class TeardownExceptionGroup(ExceptionGroup):
    """
    One or more exception occurred during teardown phase.
    """

    ...


class SkipCallbackExceptionGroup(ExceptionGroup):
    """
    One or more exception occurred during skip callback phase.
    """

    ...



from __future__ import annotations

import tarfile
from base64 import b64decode
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Protocol, TypeAlias, get_args

from ..conn import ProcessLogLevel
from .errors import ArtifactsExceptionGroup
from .misc import sanitize_path, should_collect_artifacts
from .types import MultihostOSFamily, MultihostOutcome

if TYPE_CHECKING:
    from .logging import MultihostLogger
    from .multihost import MultihostHost


# +DOCS/MultihostArtifactsType
MultihostArtifactsType: TypeAlias = Literal[
    "pytest_setup", "pytest_teardown", "topology_setup", "topology_teardown", "test"
]
"""
Multihost artifacts type.

* ``pytest_setup``: collected after :meth:`MultihostHost.pytest_setup`
* ``pytest_teardown``: collected after :meth:`MultihostHost.pytest_teardown`
* ``topology_setup``: collected after :meth:`TopologyController.topology_setup`
* ``topology_teardown``: collected after :meth:`TopologyController.topology_teardown`
* ``test``: collected after each test run
"""
# -DOCS/MultihostArtifactsType


MultihostArtifactsMode: TypeAlias = Literal["never", "on-failure", "always"]
"""
Multihost artifacts mode.

Defines if artifacts should be collected or not.
"""


class MultihostHostArtifacts(object):
    """
    Manage set of artifacts that are collected at specific places.
    """

    def __init__(self, config: list[str] | dict[str, list[str]] | None = None) -> None:
        self.pytest_setup: set[str] = set()
        """
        List of artifacts collected for host after initial pytest_setup.

        See :meth:`MultihostHost.pytest_setup`.
        """

        self.pytest_teardown: set[str] = set()
        """
        List of artifacts collected for host after final pytest_teardown.

        See :meth:`MultihostHost.pytest_teardown`.
        """

        self.test: set[str] = set()
        """
        List of artifacts collected for a test when the test run is finished.
        """

        if config is not None:
            if isinstance(config, list):
                self.test = set(config)
            elif isinstance(config, dict):
                allowed = ["pytest_setup", "pytest_teardown", "test"]
                for key in config.keys():
                    if key not in allowed:
                        raise ValueError(f"Invalid key: {key}, expected {allowed}")

                self.pytest_setup = set(config.get("pytest_setup", set()))
                self.pytest_teardown = set(config.get("pytest_teardown", set()))
                self.test = set(config.get("test", set()))
            else:
                raise TypeError(f"Unsupported type: {type(config)}, expected list or dict")

    def get(self, artifacts_type: MultihostArtifactsType) -> set[str]:
        """
        Get list of artifacts by type.

        :param artifacts_type: Type to retrieve.
        :type artifacts_type: MultihostArtifactsType
        :raises ValueError: If invalid artifacts type is given.
        :return: List of artifacts.
        :rtype: set[str]
        """
        allowed = get_args(MultihostArtifactsType)
        if artifacts_type not in allowed:
            raise ValueError(f"Invalid artifacts type {artifacts_type}, expected {allowed}")

        if not hasattr(self, artifacts_type):
            return set()

        return getattr(self, artifacts_type)


class MultihostTopologyControllerArtifacts(object):
    """
    Manage set of artifacts that are collected at specific places.
    """

    def __init__(self) -> None:
        self.topology_setup: dict[MultihostHost, set[str]] = {}
        """
        List of artifacts collected for host after initial topology_setup.

        See :meth:`TopologyController.topology_setup`.
        """

        self.topology_teardown: dict[MultihostHost, set[str]] = {}
        """
        List of artifacts collected for host after final topology_teardown.

        See :meth:`TopologyController.topology_teardown`.
        """

        self.test: dict[MultihostHost, set[str]] = {}
        """
        List of artifacts collected for host when a test run is finished.
        """

    def get(self, host: MultihostHost, artifacts_type: MultihostArtifactsType) -> set[str]:
        """
        Get list of artifacts by host and type.

        :param artifacts_type: Type to retrieve.
        :type artifacts_type: MultihostArtifactsType
        :raises ValueError: If invalid artifacts type is given.
        :return: List of artifacts.
        :rtype: set[str]
        """
        allowed = get_args(MultihostArtifactsType)
        if artifacts_type not in allowed:
            raise ValueError(f"Invalid artifacts type {artifacts_type}, expected {allowed}")

        if not hasattr(self, artifacts_type):
            return set()

        artifacts: dict[MultihostHost, set[str]] = getattr(self, artifacts_type)
        return artifacts.get(host, set())


class MultihostArtifactsCollectable(Protocol):
    """
    Protocol: object supports artifacts collection.
    """

    def get_artifacts_list(self, host: MultihostHost, artifacts_type: MultihostArtifactsType) -> set[str]:
        """
        Return the list of artifacts to collect.

        This just returns :attr:`artifacts`, but it is possible to override this
        method in order to generate additional artifacts that were not created
        by the test, or detect which artifacts were created and update the
        artifacts list.

        :param host: Host where the artifacts are being collected.
        :type host: MultihostHost
        :param artifacts_type: Type of artifacts that are being collected.
        :type artifacts_type: MultihostArtifactsType
        :return: List of artifacts to collect.
        :rtype: set[str]
        """
        pass


class MultihostArtifactsCollector(object):
    """
    Multihost artifacts collector.
    """

    def __init__(
        self,
        *,
        host: MultihostHost,
        path: Path | str,
        mode: MultihostArtifactsMode,
        compress: bool,
    ) -> None:
        """
        :param host: MultihostHost object.
        :type host: MultihostHost
        :param path: Artifacts directory path.
        :type path: Path | str
        :param mode: Artifacts collection mode.
        :type mode: MultihostArtifactsMode
        :param compress: Store artifacts in compressed tarball or not?
        :type compress: bool
        """
        self.path: Path = Path(path)
        self.host: MultihostHost = host
        self.logger: MultihostLogger = host.logger
        self.mode: MultihostArtifactsMode = mode
        self.compress: bool = compress

    def should_collect(self, outcome: MultihostOutcome) -> bool:
        """
        Match mode and outcome in order to decide if artifacts should be
        collected/written or not.

        :param outcome: Test or operation outcome.
        :type outcome: MultihostOutcome
        :raises ValueError: If mode is not recognized.
        :return: True if artifacts should be collected, False otherwise.
        :rtype: bool
        """
        return should_collect_artifacts(self.mode, outcome)

    def collect(
        self,
        artifacts_type: MultihostArtifactsType,
        *,
        path: str,
        outcome: MultihostOutcome,
        collect_objects: list[MultihostArtifactsCollectable],
    ) -> None:
        """
        Collect artifacts to $artifacts_dir/$path/$collection_path.

        :param artifacts_type: Artifacts type.
        :type artifacts_type: MultihostArtifactsType
        :param path: Artifacts path relative to artifacts directory.
        :type path: str
        :param outcome: Test or operation outcome.
        :type outcome: MultihostOutcome
        :param collect_objects: Objects from which artifacts will be collected.
        :type collect_objects: list[MultihostArtifactCollectionType]
        :raises Exception: If error happens when obtaining artifacts list.
        :raises NotImplementedError: If an attempt to collect artifacts on Windows host is performed.
        :raises ValueError: If host with unknown operating system is given.
        """
        if not self.should_collect(outcome):
            self.logger.info("Artifacts are not collected")
            return

        # Substitute problematic characters and create destination path
        dest = self.path / sanitize_path(path)

        # Gather list of artifacts to collect
        errors = []
        artifacts_set: set[str] = set()
        for obj in collect_objects:
            try:
                artifacts_set.update(obj.get_artifacts_list(self.host, artifacts_type))
            except Exception as e:
                errors.append(e)

        if errors:
            raise ArtifactsExceptionGroup("Unable to collect artifacts from all hosts", errors)

        # Sort artifacts by name
        artifacts = sorted(artifacts_set)
        if not artifacts:
            self.logger.info("No artifacts to collect.")
            return

        self.logger.info(
            "Collecting artifacts",
            extra={
                "data": {
                    "Local destination": dest if not self.compress else f"{dest}.tgz",
                    "Artifacts": artifacts,
                }
            },
        )

        # Create output directory, skip the last part since it is the
        # tarball/collection name.
        dest.parent.mkdir(parents=True, exist_ok=True)

        # Collect artifacts
        match self.host.os_family:
            case MultihostOSFamily.Linux:
                command = f"""
                    tmp=`mktemp /tmp/mh.host.artifacts.XXXXXXXXX`
                    tar -hczvf "$tmp" {' '.join([f'$(compgen -G "{x}")' for x in artifacts])} &> /dev/null
                    base64 "$tmp"
                    rm -f "$tmp" &> /dev/null
                """
            case MultihostOSFamily.Windows:
                raise NotImplementedError("Artifacts are not supported on Windows machine")
            case _:
                raise ValueError(f"Unknown operating system: {self.host.os_family}")

        result = self.host.conn.run(command, log_level=ProcessLogLevel.Error)

        # Return if no artifacts were obtained
        if not result.stdout:
            return

        with BytesIO(b64decode(result.stdout)) as buffer:
            if self.compress:
                # Store artifacts in single archive
                with open(f"{dest}.tgz", "wb") as f:
                    f.write(buffer.getbuffer())
            else:
                # Extract archive for convenience
                with tarfile.open(fileobj=buffer) as tar:
                    tar.extractall(str(dest))



from __future__ import annotations

import signal
import subprocess
import threading
from typing import IO, TYPE_CHECKING, Any, Generator, Self

import colorama as c

from pytest_mh.conn import Process, ProcessLogLevel

from .._private.logging import MultihostLogger
from . import Connection, ConnectionError, ProcessError, ProcessInputBuffer, ProcessResult, ProcessTimeoutError, Shell

if TYPE_CHECKING:
    from .. import MultihostHost

__all__ = [
    "ContainerClient",
    "ContainerConnectionError",
    "ContainerProcess",
    "ContainerProcessError",
    "ContainerProcessTimeoutError",
    "ContainerInputBuffer",
    "ContainerProcessResult",
]


class ContainerInputBuffer(ProcessInputBuffer):
    """
    Container Input Buffer.

    Allows to write into stdin of opened Container channel.
    """

    def __init__(self, pipe: IO[bytes]) -> None:
        """
        :param pipe: Input pipe.
        :type pipe: BufferedWriter
        """
        self.pipe: IO[bytes] = pipe

    def write(self, data: str | bytes) -> None:
        if isinstance(data, str):
            data = data.encode("utf-8")

        self.pipe.write(data)

    def flush(self) -> None:
        """
        Flush the input stream.
        """
        self.pipe.flush()

    def close(self) -> None:
        """
        Close the input stream.
        """
        self.pipe.close()


class ContainerOutputBuffer(Generator):
    """
    Container Output Buffer.

    Reads from stdout or stderr from the running process and makes each line
    of the data accessible through a generator.
    """

    def __init__(self, pipe: IO[bytes]):
        """
        :param pipe: Input pipe.
        :type pipe: BufferedWriter
        """
        self.pipe: IO[bytes] = pipe

        self.eof: bool = False
        self.lines: list[str] = []

        # If a buffer is full the process is paused waiting for the buffer to be
        # read. This can cause deadlock under certain situations, therefore we
        # need to keep reading the buffer continuously in another thread.
        self._index: int = 0
        self._lock: threading.Condition = threading.Condition()
        self._thread = threading.Thread(target=self._read)
        self._thread.daemon = True
        self._thread.start()

    def _read(self) -> None:
        while not self.eof:
            with self._lock:
                line: bytes = self.pipe.readline()
                if line:
                    self.lines.append(line.decode("utf-8").rstrip("\n"))
                else:
                    self.eof = True

                self._lock.notify()

    def finish(self) -> None:
        """
        Read all remaining data.
        """
        self._thread.join()

    def send(self, value: Any):
        with self._lock:
            self._lock.wait_for(lambda: self.eof or self._index < len(self.lines))
            if self.eof and self._index >= len(self.lines):
                raise StopIteration

            line: str = self.lines[self._index]
            self._index += 1
            return line

    def throw(self, typ, val=None, tb=None):
        super().throw(typ, val, tb)


class ContainerProcessError(ProcessError):
    """
    Container Process Error.
    """

    pass


class ContainerProcessTimeoutError(ProcessTimeoutError):
    """
    Container Process Timeout Error.
    """

    pass


class ContainerProcessResult(ProcessResult[ContainerProcessError]):
    """
    Container Process result.
    """

    pass


class ContainerProcess(Process[ContainerProcessResult, ContainerInputBuffer, ContainerProcessTimeoutError]):
    """
    Container Process manager.
    """

    def __init__(
        self,
        *,
        command: str,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | bytes | None = None,
        shell: Shell,
        logger: MultihostLogger,
        log_level: ProcessLogLevel,
        timeout: int,
        blocking_call: bool,
        client: ContainerClient,
    ) -> None:
        """
        :param command: Command to execute.
        :type command: str
        :param cwd: Working directory, defaults to None
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | bytes | None, optional
        :param shell: Shell used to execute the command, defaults to None (use user's login shell)
        :type shell: str | None, optional
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param log_level: Log level.
        :type log_level: ProcessLogLevel
        :param timeout: Timeout in seconds, value ``0`` means that timeout is
            disabled.
        :type timeout: int
        :param blocking_call: Is this a blocking execution?
        :type blocking_call: bool
        :param client: Container client.
        :type client: ContainerClient
        """
        super().__init__(
            command=command,
            cwd=cwd,
            env=env,
            input=input,
            shell=shell,
            logger=logger,
            log_level=log_level,
            timeout=timeout,
            blocking_call=blocking_call,
            additional_log_data={
                "Engine": client.engine,
                "Container": client.container_name,
                "User": client.user,
            },
        )

        self.__client: ContainerClient = client

        self.__stdout: ContainerOutputBuffer | None = None
        self.__stderr: ContainerOutputBuffer | None = None
        self.__stdin: ContainerInputBuffer | None = None
        self.__popen: subprocess.Popen | None = None

    @property
    def in_progress(self) -> bool:
        return self.__popen is not None

    @property
    def stdout(self) -> Generator[str, None, None]:
        if not self.in_progress or self.__stdout is None:
            raise RuntimeError("Accessing stdout on a process that is not running.")

        return self.__stdout

    @property
    def stderr(self) -> Generator[str, None, None]:
        if not self.in_progress or self.__stderr is None:
            raise RuntimeError("Accessing stderr on a process that is not running.")

        return self.__stderr

    @property
    def stdin(self) -> ContainerInputBuffer:
        if not self.in_progress or self.__stdin is None:
            raise RuntimeError("Accessing stdin on a process that is not running.")

        return self.__stdin

    def _run(self) -> None:
        """
        Execute the command.

        This is an internal method called by :meth:`run` after executing
        generic code.
        """
        command = "{sudo} {engine} exec --interactive {name} {command}".format(
            sudo="sudo -k -S --prompt=''" if self.__client.sudo else "",
            engine=self.__client.engine,
            name=self.__client.container_name,
            command=self.full_command_line,
        ).lstrip()

        try:
            self.__popen = subprocess.Popen(
                args=command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )

            # This is to satisfy mypy
            if self.__popen.stdout is None or self.__popen.stderr is None or self.__popen.stdin is None:
                raise RuntimeError("subprocess.Popen did not correctly open pipes")

            self.__stdin = ContainerInputBuffer(self.__popen.stdin)
            self.__stdout = ContainerOutputBuffer(self.__popen.stdout)
            self.__stderr = ContainerOutputBuffer(self.__popen.stderr)

            if self.__client.sudo and self.__client.sudo_password is not None:
                self.stdin.write(f"{self.__client.sudo_password}\n")
                self.stdin.flush()

            if self.input is not None:
                self.stdin.write(self.input)
                self.stdin.flush()

        except Exception:
            self._close()
            raise

    def _wait(self) -> ContainerProcessResult:
        """
        Wait for the command to finish.

        EOF is send to standard input to indicate that there will be no
        additional input data. Then it waits for the command to finish.

        This is an internal method called by :meth:`run` after executing
        generic code.

        :return: Command result.
        :rtype: ContainerProcessResult
        """
        if self.__popen is None or self.__stdout is None or self.__stderr is None:
            raise RuntimeError("Calling wait on process that has not yet started.")

        try:
            # Notify the program that there will be no more input
            self.send_eof()

            # Read remaining output, this will finish the output generator and append
            # remaining lines to self.__stdout and self.__stderr buffers.
            self.__stdout.finish()
            self.__stderr.finish()

            # Wait for the program to finish and get the exit code.
            code = self.__popen.wait()

            error = ContainerProcessError(
                code, self.id, self.command, self.cwd, self.env, self.input, self.__stdout.lines, self.__stderr.lines
            )

            result = ContainerProcessResult(code, self.__stdout.lines, self.__stderr.lines, error)
        except TimeoutError as e:
            raise ContainerProcessTimeoutError(
                e.args[0],
                self.id,
                self.command,
                self.cwd,
                self.env,
                self.input,
                self.__stdout.lines,
                self.__stderr.lines,
            )
        finally:
            self._close()

        return result

    def send_eof(self) -> None:
        if self.__popen is None or self.__stdin is None:
            raise RuntimeError("Calling send_eof on process that is not running.")

        self.__stdin.close()

    def send_signal(self, sig: signal.Signals) -> None:
        if self.__popen is None:
            raise RuntimeError("Calling send_signal on process that is not running.")

        self.__popen.send_signal(sig)

    def _close(self) -> None:
        if self.__popen is None:
            return

        if self.__popen.returncode is None:
            self.__popen.kill()

        self.__popen = None
        self.__stdout = None
        self.__stderr = None
        self.__stdin = None


class ContainerConnectionError(ConnectionError):
    """
    Unable to connect to the container.
    """

    def __init__(
        self,
        engine: str,
        container_name: str,
        *,
        user: str,
        sudo: bool,
    ) -> None:
        super().__init__(f"Unable to connect to {engine} container {container_name}, user={user}, sudo={sudo}")


class ContainerClient(Connection[ContainerProcess, ContainerProcessResult]):
    """
    Interactive podman and docker client.
    """

    def __init__(
        self,
        engine: str,
        container_name: str,
        *,
        user: str,
        sudo: bool = False,
        sudo_password: str | None = None,
        shell: Shell,
        logger: MultihostLogger,
        timeout: int = 300,
    ) -> None:
        """
        :param container_name: Container name.
        :type container_name: str
        :param user: Username that will be used to execute commands.
        :type user: str
        :param sudo: Run podman under root, defaults to ``False``.
        :type sudo: bool
        :param sudo_password: SUDO password, defaults to ``None``.
        :type sudo_password: str | None
        :param shell: User shell used to run commands, defaults to ContainerBashProcess
        :type shell: str, optional
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param timeout: Timeout in seconds (defaults to 300), value
            ``0`` means that timeout is disabled.
        :type timeout: int
        """
        super().__init__(shell=shell, logger=logger, timeout=timeout)

        if engine not in ("podman", "docker"):
            raise ValueError(f"Unsupported container engine {engine}, expected podman or docker!")

        self.engine: str = engine
        self.container_name: str = container_name
        self.user: str = user

        self.sudo: bool = sudo
        self.sudo_password: str | None = sudo_password

        self._connected: bool = False

    @property
    def connected(self) -> bool:
        return self._connected

    def connect(self) -> None:
        """
        Connect to the host.

        :raises ContainerAuthenticationError: If user fails to authenticate.
        """
        if self.connected:
            return

        self.logger.info(
            self.logger.colorize("Checking container ", c.Style.BRIGHT)
            + self.logger.colorize(self.container_name, c.Fore.BLUE, c.Style.BRIGHT)
            + self.logger.colorize(f" using {self.engine}", c.Style.BRIGHT)
        )

        # We need to mark it as connected here to avoid recursion from `run`
        self._connected = True

        # Check that the container is reachable
        result = self.run("exit 0", raise_on_error=False, log_level=ProcessLogLevel.Error)
        if result.rc != 0:
            self._connected = False
            raise ContainerConnectionError(self.engine, self.container_name, user=self.user, sudo=self.sudo)

    def disconnect(self) -> None:
        self._connected = False

    def create_process(
        self,
        *,
        command: str,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | bytes | None = None,
        log_level: ProcessLogLevel,
        blocking_call: bool,
    ) -> ContainerProcess:
        return ContainerProcess(
            command=command,
            cwd=cwd,
            env=env,
            input=input,
            shell=self.shell,
            logger=self.logger,
            log_level=log_level,
            timeout=self.timeout,
            blocking_call=blocking_call,
            client=self,
        )

    @classmethod
    def from_confdict(cls, host: MultihostHost, confdict: dict[str, Any]) -> Self:
        engine: str = confdict["type"]
        container: str | None = confdict.get("container", None)
        user: str = confdict.get("user", "root")
        sudo: bool = confdict.get("sudo", False)
        sudo_password: str | None = confdict.get("sudo_password", None)
        timeout: int = confdict.get("timeout", 300)

        if container is None:
            raise ValueError("Container name is not set!")

        return cls(
            engine=engine,
            container_name=container,
            user=user,
            sudo=sudo,
            sudo_password=sudo_password,
            logger=host.logger,
            shell=host.shell,
            timeout=timeout,
        )



from __future__ import annotations

from collections import deque
from typing import Self

from .. import MultihostHost, MultihostReentrantUtility
from ..conn import Process, ProcessLogLevel, ProcessResult

__all__ = ["SystemdServices"]


class SystemdServices(MultihostReentrantUtility[MultihostHost]):
    """
    Manage remote services.
    """

    def __init__(self, host: MultihostHost) -> None:
        super().__init__(host)
        self.initial_states: dict[str, bool] = {}
        self.__states: deque[dict[str, bool]] = deque()

    def __enter__(self) -> Self:
        """
        Saves current state.

        :return: Self.
        :rtype: Self
        """
        self.__states.append(self.initial_states)
        self.initial_states = {}

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Revert all changes done during current context.
        """
        # Restart all services that were touched
        self.reload_daemon()
        for service, state in self.initial_states.items():
            self.logger.info(f'systemd: restoring "{service}" to {"started" if state else "stopped"}')
            self.host.conn.run(
                self._debug_failure(service, f'systemctl stop "{service}"'),
                raise_on_error=False,
                log_level=ProcessLogLevel.Error,
            )
            if state:
                self.host.conn.run(
                    self._debug_failure(service, f'systemctl start "{service}"'),
                    raise_on_error=False,
                    log_level=ProcessLogLevel.Error,
                )

        self.initial_states = self.__states.pop()

    def async_start(self, service: str) -> Process:
        """
        Start a systemd unit. Non-blocking call.

        ``systemctl status $unit`` is called automatically if the unit can not
        be started. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: Process
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: starting "{service}" asynchronously')
        return self.host.conn.async_run(
            self._debug_failure(service, f'systemctl reset-failed "{service}"; systemctl start "{service}"'),
            log_level=ProcessLogLevel.Error,
        )

    def start(self, service: str, raise_on_error: bool = True) -> ProcessResult:
        """
        Start a systemd unit. The call will wait until the operation is finished.

        ``systemctl status $unit`` is called automatically if the unit can not
        be started. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: ProcessResult
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: starting "{service}"')
        return self.host.conn.run(
            self._debug_failure(service, f'systemctl reset-failed "{service}"; systemctl start "{service}"'),
            raise_on_error=raise_on_error,
            log_level=ProcessLogLevel.Error,
        )

    def async_stop(self, service: str) -> Process:
        """
        Stop a systemd unit. Non-blocking call.

        ``systemctl status $unit`` is called automatically if the unit can not
        be stopped. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: Process
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: stopping "{service}" asynchronously')
        return self.host.conn.async_run(
            self._debug_failure(service, f'systemctl stop "{service}"'), log_level=ProcessLogLevel.Error
        )

    def stop(self, service: str, raise_on_error: bool = True) -> ProcessResult:
        """
        Stop a systemd unit. The call will wait until the operation is finished.

        ``systemctl status $unit`` is called automatically if the unit can not
        be stoped. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: ProcessResult
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: stopping "{service}"')
        return self.host.conn.run(
            self._debug_failure(service, f'systemctl stop "{service}"'),
            raise_on_error=raise_on_error,
            log_level=ProcessLogLevel.Error,
        )

    def async_restart(self, service: str) -> Process:
        """
        Restart a systemd unit. Non-blocking call.

        ``systemctl status $unit`` is called automatically if the unit can not
        be restarted. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: Process
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: restarting "{service}" asynchronously')
        return self.host.conn.async_run(
            self._debug_failure(service, f'systemctl reset-failed "{service}"; systemctl restart "{service}"'),
            log_level=ProcessLogLevel.Error,
        )

    def restart(self, service: str, raise_on_error: bool = True) -> ProcessResult:
        """
        Restart a systemd unit. The call will wait until the operation is finished.

        ``systemctl status $unit`` is called automatically if the unit can not
        be restarted. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: ProcessResult
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: restarting "{service}"')
        return self.host.conn.run(
            self._debug_failure(service, f'systemctl reset-failed "{service}"; systemctl restart "{service}"'),
            raise_on_error=raise_on_error,
            log_level=ProcessLogLevel.Error,
        )

    def async_reload(self, service: str) -> Process:
        """
        Reload a systemd unit. Non-blocking call.

        ``systemctl status $unit`` is called automatically if the unit can not
        be reloaded. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: Process
        """
        self.logger.info(f'systemd: reloading "{service}" asynchronously')
        return self.host.conn.async_run(
            self._debug_failure(service, f'systemctl reload "{service}"'), log_level=ProcessLogLevel.Error
        )

    def reload(self, service: str, raise_on_error: bool = True) -> ProcessResult:
        """
        Reload a systemd unit. The call will wait until the operation is finished.

        ``systemctl status $unit`` is called automatically if the unit can not
        be reloaded. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: ProcessResult
        """
        self.logger.info(f'systemd: reloading "{service}"')
        return self.host.conn.run(
            self._debug_failure(service, f'systemctl reload "{service}"'),
            raise_on_error=raise_on_error,
            log_level=ProcessLogLevel.Error,
        )

    def async_status(self, service: str) -> Process:
        """
        Get systemd unit status. Non-blocking call.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: Process
        """
        return self.host.conn.async_run(f'systemctl status "{service}"')

    def status(self, service: str, raise_on_error: bool = True) -> ProcessResult:
        """
        Get systemd unit status. The call will wait until the operation is finished.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: ProcessResult
        """
        return self.host.conn.run(f'systemctl status "{service}"', raise_on_error=raise_on_error)

    def async_get_property(self, service: str, prop: str) -> Process:
        """
        Get property of systemd unit. Non-blocking call.

        :param service: Unit name.
        :type service: str
        :param prop: Propery name.
        :type prop: str
        :return: Running SSH process.
        :rtype: Process
        """
        return self.host.conn.async_run(f'systemctl show "{service}" --value --property "{prop}"')

    def get_property(self, service: str, prop: str, raise_on_error: bool = True) -> str:
        """
        Get property of systemd unit. The call will wait until the operation is finished.

        :param service: Unit name.
        :type service: str
        :param prop: Propery name.
        :type prop: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: property value as string.
        :rtype: str
        """
        result = self.host.conn.run(
            f'systemctl show "{service}" --value --property "{prop}"', raise_on_error=raise_on_error
        )
        return result.stdout.strip()

    def async_reload_daemon(self) -> Process:
        """
        Reload systemd daemon to refresh unit files. Non-blocking call.

        :return: Running SSH process.
        :rtype: Process
        """
        self.logger.info("systemd: reloading systemd daemon")
        return self.host.conn.async_run("systemctl daemon-reload", log_level=ProcessLogLevel.Error)

    def reload_daemon(self, raise_on_error: bool = True) -> ProcessResult:
        """
        Reload systemd daemon to refresh unit files. The call will wait until the operation is finished.

        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: ProcessResult
        """
        return self.host.conn.run("systemctl daemon-reload", raise_on_error=raise_on_error)

    def _debug_failure(self, service: str, command: str) -> str:
        return f"{command} || ({self._query_logs_command(service)})"

    def _query_logs_command(self, service: str) -> str:
        cmds = [
            "rc=$?",
            f"echo '+ systemctl status {service}'",
            f"systemctl status {service}",
            "echo ''",
            f"echo '+ journalctl --no-pager -xeu {service}'",
            f"journalctl --no-pager -xeu {service}",
            "exit $rc",
        ]

        return ";".join(cmds)

    def __set_initial_state(self, service: str) -> None:
        if service in self.initial_states:
            return

        result = self.host.conn.run(
            f'systemctl status "{service}"', log_level=ProcessLogLevel.Silent, raise_on_error=False
        )

        if result.rc == 0 or (result.rc == 3 and "Active: activating" in result.stdout):
            self.initial_states[service] = True
        else:
            self.initial_states[service] = False

    def is_active(self, service: str) -> bool:
        """
        Check if a systemd unit is active.

        :param service: Unit name.
        :type service: str
        :return: True if unit is active, False otherwise.
        :rtype: bool
        """
        result = self.host.conn.run(
            f'systemctl is-active "{service}"', raise_on_error=False, log_level=ProcessLogLevel.Silent
        )
        return result.rc == 0



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace
from typing import Any, Callable, Generator

import pytest

from .artifacts import MultihostArtifactsCollectable
from .data import MultihostItemData
from .errors import SkipCallbackExceptionGroup, TeardownExceptionGroup
from .logging import MultihostLogger
from .marks import TopologyMark
from .misc import invoke_callback
from .multihost import (
    MultihostConfig,
    MultihostDomain,
    MultihostHost,
    MultihostRole,
    mh_utility_enter_dependencies,
    mh_utility_exit_dependencies,
    mh_utility_pytest_report_teststatus,
    mh_utility_setup_dependencies,
    mh_utility_teardown_dependencies,
)
from .topology import Topology, TopologyDomain
from .topology_controller import TopologyController


class MultihostFixture(object):
    """
    Multihost object provides access to underlaying multihost configuration,
    individual domains and hosts. This object should be used only in tests
    as the :func:`mh` pytest fixture.

    Domains are accessible as dynamically created properties of this object,
    hosts are accessible by roles as dynamically created properties of each
    domain. Each host object is instance of specific role class based on
    :mod:`~pytest_mh.MultihostRole`.

    .. code-block:: yaml
        :caption: Example multihost configuration

        domains:
        - id: test
          hosts:
          - name: client
            hostname: client.test
            role: client

          - name: ldap
            hostname: master.ldap.test
            role: ldap

    The configuration above creates one domain of id ``test`` with two hosts.
    The following example shows how to access the hosts:

    .. code-block:: python
        :caption: Example of the MultihostFixture object

        def test_example(mh: MultihostFixture):
            mh.ns.test            # -> namespace containing roles as properties
            mh.ns.test.client     # -> list of hosts providing given role
            mh.ns.test.client[0]  # -> host object, instance of specific role
    """

    def __init__(
        self,
        request: pytest.FixtureRequest,
        data: MultihostItemData,
        multihost: MultihostConfig,
        topology_mark: TopologyMark,
    ) -> None:
        """
        :param request: Pytest request.
        :type request: pytest.FixtureRequest
        :param data: Multihost item data.
        :type data: MultihostItemData
        :param multihost: Multihost configuration.
        :type multihost: MultihostConfig
        :param topology_mark: Multihost topology mark.
        :type topology_mark: TopologyMark
        """

        self.data: MultihostItemData = data
        """
        Multihost item data.
        """

        self.request: pytest.FixtureRequest = request
        """
        Pytest request.
        """

        self.multihost: MultihostConfig = multihost
        """
        Multihost configuration.
        """

        self.topology_mark: TopologyMark = topology_mark
        """
        Topology mark.
        """

        self.topology: Topology = topology_mark.topology
        """
        Topology data.
        """

        self.topology_controller: TopologyController = topology_mark.controller
        """
        Topology controller.
        """

        self.logger: MultihostLogger = multihost.logger
        """
        Multihost logger.
        """

        self.roles: list[MultihostRole] = []
        """
        Available MultihostRole objects.
        """

        self.hosts: list[MultihostHost] = []
        """
        Available MultihostHost objects.
        """

        self.fixtures: dict[str, MultihostRole | list[MultihostRole]] = {}
        """
        All dynamic fixtures defined in the topology mapped from name to :class:`MultihostRole`.
        """

        self.ns: SimpleNamespace = SimpleNamespace()
        """
        Roles as object accessible through topology path, e.g. ``mh.ns.domain_id.role_name``.
        """

        self._paths: dict[str, list[MultihostRole] | MultihostRole] = {}
        self._skipped: bool = False

        for domain in self.multihost.domains:
            if domain.id in self.topology:
                setattr(self.ns, domain.id, self._domain_to_namespace(domain, self.topology.get(domain.id)))

        self.roles = sorted([x for x in self._paths.values() if isinstance(x, MultihostRole)], key=lambda x: x.role)
        self.hosts = sorted(list({x.host for x in self.roles}), key=lambda x: x.hostname)
        self.fixtures = self.topology_mark.map_fixtures_to_roles(self)

    def _domain_to_namespace(self, domain: MultihostDomain, topology_domain: TopologyDomain) -> SimpleNamespace:
        ns = SimpleNamespace()
        for role_name in domain.roles:
            if role_name not in topology_domain:
                continue

            count = topology_domain.get(role_name)
            roles = [domain.create_role(self, host) for host in domain.hosts_by_role(role_name)[:count]]

            self._paths[f"{domain.id}.{role_name}"] = roles
            for index, role in enumerate(roles):
                self._paths[f"{domain.id}.{role_name}[{index}]"] = role

            setattr(ns, role_name, roles)

        return ns

    def _lookup(self, path: str) -> MultihostRole | list[MultihostRole]:
        """
        Lookup host by path. The path format is ``$domain.$role``
        or ``$domain.$role[$index]``

        :param path: Host path.
        :type path: str
        :raises LookupError: If host is not found.
        :return: The role object if index was given, list of role objects otherwise.
        :rtype: MultihostRole | list[MultihostRole]
        """

        if path not in self._paths:
            raise LookupError(f'Name "{path}" does not exist')

        return self._paths[path]

    def _skip(self) -> bool:
        self._skipped = False

        try:
            try:
                self.log_phase("SKIP BY TOPOLOGY")
                reason = self._skip_by_topology(self.topology_controller)
                if reason is not None:
                    self._skipped = True
                    pytest.skip(reason)
            finally:
                self.log_phase("SKIP BY TOPOLOGY DONE")

            try:
                self.log_phase("SKIP BY REQUIRE MARKER")
                reason = self._skip_by_require_marker(self.topology_mark, self.request.node)
                if reason is not None:
                    self._skipped = True
                    pytest.skip(reason)
            finally:
                self.log_phase("SKIP BY REQUIRE MARKER DONE")
        except Exception as e:
            # Just re-raise if pytest.skip was called, unfortunately pytest.skip
            # raises exception that is not publicly available, so we have to
            # rely on our attribute.
            if self._skipped:
                raise

            # Error out and skip the test
            self.data.outcome = "error"
            self._skipped = True
            raise SkipCallbackExceptionGroup("An exception occurred inside a skip callback", [e])
        finally:
            self.split_log_file("skip.log")
            if self.data.outcome == "error":
                self.logger.flush(self.data.outcome)

        return self._skipped

    def _skip_by_topology(self, controller: TopologyController):
        return controller._invoke_with_args(controller.skip)

    def _skip_by_require_marker(self, topology_mark: TopologyMark, node: pytest.Function) -> str | None:
        fixtures: dict[str, Any] = {k: None for k in topology_mark.fixtures.keys()}
        fixtures.update(node.funcargs)
        topology_mark.apply(self, fixtures)

        # Make sure mh fixture is always available
        fixtures["mh"] = self

        for mark in node.iter_markers("require"):
            if len(mark.args) not in [1, 2]:
                raise ValueError(f"{node.nodeid}::{node.originalname}: " "invalid arguments for @pytest.mark.require")

            condition = mark.args[0]
            reason = "Required condition was not met" if len(mark.args) != 2 else mark.args[1]

            callresult = invoke_callback(condition, **fixtures)
            if isinstance(callresult, tuple):
                if len(callresult) != 2:
                    raise ValueError(
                        f"{node.nodeid}::{node.originalname}: " "invalid arguments for @pytest.mark.require"
                    )

                result = callresult[0]
                reason = callresult[1]
            else:
                result = callresult

            if not result:
                return reason

        return None

    def _setup_hosts_utils(self) -> None:
        """
        Enter reentrant utilities in each host.
        """
        for item in self.hosts:
            mh_utility_enter_dependencies(item, "test")

    def _setup_hosts(self) -> None:
        """
        Run per-test setup of each host.
        """
        for item in self.hosts:
            item._op_state.clear("setup")

        for item in self.hosts:
            item.setup()
            item._op_state.set_success("setup")

    def _setup_topology(self) -> None:
        """
        Run per-test setup of topology controller.
        """
        self.topology_controller._invoke_with_args(self.topology_controller.setup)
        self.topology_controller._op_state.set_success("setup")

    def _setup_utils(self) -> None:
        """
        Run utility setup for each role.
        """
        for item in self.roles:
            mh_utility_setup_dependencies(item)

    def _setup_roles(self) -> None:
        """
        Run per-test setup of each role.
        """
        for item in self.roles:
            item.setup()
            item._op_state.set_success("setup")

    def _teardown_roles(self) -> None:
        """
        Run per-test teardown of each role.
        """
        errors = []
        for item in self.roles:
            if item._op_state.check_success("setup"):
                try:
                    item.teardown()
                except Exception as e:
                    errors.append(e)

        if errors:
            raise TeardownExceptionGroup("Unable to teardown some roles (role.teardown)", errors)

    def _teardown_utils(self) -> None:
        """
        Run utility teardown for each role.
        """
        for item in self.roles:
            mh_utility_teardown_dependencies(item)

    def _teardown_topology(self) -> None:
        """
        Run per-test teardown from topology controller.
        """
        if self.topology_controller._op_state.check_success("setup"):
            self.topology_controller._invoke_with_args(self.topology_controller.teardown)

    def _teardown_hosts(self) -> None:
        """
        Run per-test teardown of each host.
        """
        errors = []
        for item in self.hosts:
            if item._op_state.check_success("setup"):
                try:
                    item.teardown()
                except Exception as e:
                    errors.append(e)

        if errors:
            raise TeardownExceptionGroup("Unable to teardown some hosts (host.teardown)", errors)

    def _teardown_hosts_utils(self) -> None:
        """
        Exit reentrant utilities in each host.
        """
        errors = []
        for item in self.hosts:
            try:
                mh_utility_exit_dependencies(item, "test")
            except Exception as e:
                errors.append(e)

        if errors:
            raise TeardownExceptionGroup("Unable to exit some utilities (util.__exit__)", errors)

    def _pytest_report_teststatus(
        self, report: pytest.CollectReport | pytest.TestReport, config: pytest.Config
    ) -> tuple[str, str, str | tuple[str, dict[str, bool]]] | None:
        """
        Run pytest_report_teststatus on each utility.
        """
        for item in self.roles + self.hosts:
            result = mh_utility_pytest_report_teststatus(item, report, config)
            if result is not None:
                # Change stored outcome since the hook may have changed it.
                self.data.outcome = report.outcome
                return result

        return None

    def _collect_artifacts(self) -> None:
        # Create list of collectable objects
        collectable: dict[MultihostHost, list[MultihostArtifactsCollectable]] = {}
        for role in self.roles:
            host_collection = collectable.setdefault(role.host, [role.host, self.topology_controller, role])
            host_collection.extend(role._mh_utility_dependencies)

        # Collect artifacts, if an error is raised, we will ignore it since
        # teardown is more important
        for host in self.hosts:
            try:
                host.artifacts_collector.collect(
                    "test",
                    path=f"tests/{self.request.node.name}/{host.role}/{host.hostname}",
                    outcome=self.data.outcome,
                    collect_objects=collectable[host],
                )
            except Exception as e:
                self.logger.error(
                    "An error happend when collecting artifacts",
                    extra={
                        "data": {
                            "Error message": str(e),
                        }
                    },
                )

    def split_log_file(self, name: str) -> None:
        """
        Split current log records into a log file.

        :param name: Log file name.
        :type name: str
        """
        self.logger.split(Path(f"tests/{self.request.node.name}") / name)

    def _invoke_phase(self, name: str, cb: Callable, catch: bool = False) -> Exception | None:
        self.log_phase(name)
        try:
            cb()
        except Exception as e:
            if catch:
                return e

            raise
        finally:
            self.log_phase(f"{name} DONE")

        return None

    def log_phase(self, phase: str) -> None:
        """
        Log current test phase.

        :param phase: Phase name or description.
        :type phase: str
        """
        self.logger.phase(f"{phase} :: {self.request.node.nodeid}")

    def _enter(self) -> MultihostFixture:
        if self._skip():
            return self

        try:
            self._invoke_phase("SETUP ENTER HOSTS UTILS", self._setup_hosts_utils)
            self._invoke_phase("SETUP HOSTS", self._setup_hosts)
            self._invoke_phase("SETUP TOPOLOGY", self._setup_topology)
            self._invoke_phase("SETUP UTILS", self._setup_utils)
            self._invoke_phase("SETUP ROLES", self._setup_roles)
        except Exception:
            self.data.outcome = "error"
            raise
        finally:
            self.split_log_file("setup.log")

        return self

    def _exit(self) -> None:
        if self._skipped:
            return

        errors: list[Exception | None] = []
        errors.append(self._invoke_phase("COLLECT ARTIFACTS", self._collect_artifacts, catch=True))
        errors.append(self._invoke_phase("TEARDOWN ROLES", self._teardown_roles, catch=True))
        errors.append(self._invoke_phase("TEARDOWN UTILS", self._teardown_utils, catch=True))
        errors.append(self._invoke_phase("TEARDOWN TOPOLOGY", self._teardown_topology, catch=True))
        errors.append(self._invoke_phase("TEARDOWN HOSTS", self._teardown_hosts, catch=True))
        errors.append(self._invoke_phase("TEARDOWN EXIT HOSTS UTILS", self._teardown_hosts_utils, catch=True))

        self.split_log_file("teardown.log")
        self.logger.flush(self.data.outcome)

        all_errors = [x for x in errors if x is not None]
        if all_errors:
            raise TeardownExceptionGroup("One or more error occurred during test teardown", all_errors)


@pytest.fixture(scope="function")
def mh(request: pytest.FixtureRequest) -> Generator[MultihostFixture, None, None]:
    """
    Pytest multihost fixture. Returns instance of :class:`MultihostFixture`.
    When a pytest test is finished, this fixture takes care of tearing down the
    :class:`MultihostFixture` object automatically in order to clean up after
    the test run.

    .. note::

        It is preferred that the test case does not use this fixture directly
        but rather access the hosts through dynamically created role fixtures
        that are defined in ``@pytest.mark.topology``.

    :param request: Pytest's ``request`` fixture.
    :type request: pytest.FixtureRequest
    :raises ValueError: If not multihost configuration was given.
    :yield: MultihostFixture
    """

    data: MultihostItemData | None = MultihostItemData.GetData(request.node)
    if data is None:
        nodeid = f"{request.node.parent.nodeid}::{request.node.originalname}"
        raise ValueError(f"{nodeid}: mh fixture requested but no multihost configuration was provided")

    if data.multihost is None:
        raise ValueError("data.multihost must not be None")

    if data.topology_mark is None:
        raise ValueError("data.topology_mark must not be None")

    mh = MultihostFixture(request, data, data.multihost, data.topology_mark)
    try:
        mh._enter()
        mh.log_phase("TEST")

        data.multihost._in_test = True
        if data.multihost._sigint:
            pytest.skip("SIGINT received, aborting running test.")

        yield mh

        mh.log_phase("TEST DONE")
    finally:
        data.multihost._in_test = False
        if data.outcome == "failed" and data.result is not None:
            mh.logger.error(data.result.longreprtext)

        if not mh._skipped:
            mh.split_log_file("test.log")

        mh._exit()


@pytest.fixture(scope="function")
def mh_config(mh: MultihostFixture) -> MultihostConfig:
    """
    Multihost configuration.

    :param mh: mh fixture
    :type mh: MultihostFixture
    :return: Multihost configuration
    :rtype: MultihostConfig
    """
    return mh.multihost


@pytest.fixture(scope="function")
def mh_logger(mh: MultihostFixture) -> MultihostLogger:
    """
    Multihost logger.

    Can be used to log messages into the test log.

    :param mh: mh fixture
    :type mh: MultihostFixture
    :return: Multihost logger.
    :rtype: MultihostLogger
    """
    return mh.logger


@pytest.fixture(scope="function")
def mh_topology(mh: MultihostFixture) -> Topology:
    """
    Current topology.

    :param mh: mh fixture
    :type mh: MultihostFixture
    :return: Current topology
    :rtype: Topology
    """
    return mh.topology


@pytest.fixture(scope="function")
def mh_topology_name(mh: MultihostFixture) -> str:
    """
    Current topology name.

    :param mh: mh fixture
    :type mh: MultihostFixture
    :return: Current topology name
    :rtype: str
    """
    return mh.topology_mark.name


@pytest.fixture(scope="function")
def mh_topology_mark(mh: MultihostFixture) -> TopologyMark:
    """
    Current topology mark.

    :param mh: mh fixture
    :type mh: MultihostFixture
    :return: Current topology mark
    :rtype: TopologyMark
    """
    return mh.topology_mark



from __future__ import annotations

from enum import Enum
from typing import Literal, TypeAlias


class MultihostOSFamily(Enum):
    """
    Host operating system family.
    """

    Linux = "linux"
    Windows = "windows"


MultihostOutcome: TypeAlias = Literal["passed", "failed", "skipped", "error", "unknown"]



from __future__ import annotations

import itertools
import os
import shlex
import signal
import textwrap
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, Generator, Generic, NoReturn, Self, TypeVar

import colorama as c

from .._private.logging import MultihostLogger
from .._private.misc import timeout as timeout_decorator

if TYPE_CHECKING:
    from .. import MultihostHost

__all__ = [
    "Bash",
    "Connection",
    "ConnectionError",
    "Powershell",
    "Process",
    "ProcessError",
    "ProcessErrorType",
    "ProcessTimeoutError",
    "ProcessInputBuffer",
    "ProcessInputBufferType",
    "ProcessLogLevel",
    "ProcessResult",
    "ProcessResultType",
    "ProcessType",
    "Shell",
]


class ProcessLogLevel(Enum):
    """
    Process log level.
    """

    Silent = auto()
    """
    No log messages are produced.
    """

    Short = auto()
    """
    Command execution and return code is logged. Its output is omitted.
    """

    Full = auto()
    """
    Command execution, its return code and output is logged.
    """

    Error = auto()
    """
    Only log the command and its result on non-zero exit code.
    """


class ProcessInputBuffer(ABC):
    """
    Process' input buffer.

    Allows to write into stdin of the process.
    """

    @abstractmethod
    def write(self, data: str | bytes) -> None:
        """
        Write data to stdin.

        :param data: Data to write.
        :type data: str | bytes
        """
        pass


class ProcessErrorBase(Exception):
    """
    Base class for process errors.
    """

    def __init__(
        self,
        message: str,
        id: int,
        command: str,
        cwd: str | None,
        env: dict[str, Any],
        input: str | bytes | None,
        stdout: list[str],
        stderr: list[str],
    ) -> None:
        pretty_env = ""
        for key, value in env.items():
            pretty_env += f"{key}={value}\n"

        str_stdout: str = "\n".join(stdout)
        str_stderr: str = "\n".join(stderr)

        def dumps(value) -> str:
            if not value:
                return ""

            return "\n" + textwrap.indent(value, " " * 20)

        super().__init__(
            textwrap.dedent(
                f"""
                Command #{id} {message}:
                  Command:{dumps(command)}
                  CWD:{dumps(cwd)}
                  Env:{dumps(pretty_env.strip())}
                  Output:{dumps(str_stdout)}
                  Error output:{dumps(str_stderr)}
                """
            )
        )

        self.message: str = message
        """Error message."""

        self.id: int = id
        """Command autogenerated ID."""

        self.command: str = command
        """Failed command."""

        self.cwd: str | None = cwd
        """Working directory."""

        self.env: dict[str, Any] = env
        """Additional environment variables."""

        self.input: str | bytes | None = input
        """Input data."""

        self.stdout: str = str_stdout
        """Standard output."""

        self.stderr: str = str_stderr
        """Standard error output."""

        self.stdout_lines: list[str] = stdout
        """Standard output as list of lines."""

        self.stderr_lines: list[str] = stderr
        """Standard error output as list of lines."""


class ProcessError(ProcessErrorBase):
    """
    Process Error.
    """

    def __init__(
        self,
        rc: int,
        id: int,
        command: str,
        cwd: str | None,
        env: dict[str, Any],
        input: str | bytes | None,
        stdout: list[str],
        stderr: list[str],
    ) -> None:
        super().__init__(f"exited with return code {rc}", id, command, cwd, env, input, stdout, stderr)

        self.rc: int = rc
        """Return code."""


ProcessErrorType = TypeVar("ProcessErrorType", bound=ProcessError)
"""Generic process error type. Must be a subclass of :class:`ProcessError`."""


class ProcessTimeoutError(ProcessErrorBase):
    """
    Process Timeout Error.
    """

    def __init__(
        self,
        timeout: int,
        id: int,
        command: str,
        cwd: str | None,
        env: dict[str, Any],
        input: str | bytes | None,
        stdout: list[str],
        stderr: list[str],
    ) -> None:
        super().__init__(f"timed out after {timeout} seconds", id, command, cwd, env, input, stdout, stderr)

        self.timeout: int = timeout
        """Number of seconds required for time out."""


ProcessTimeoutErrorType = TypeVar("ProcessTimeoutErrorType", bound=ProcessTimeoutError)
"""Generic process error type. Must be a subclass of :class:`ProcessTimeoutError`."""


class ProcessResult(Generic[ProcessErrorType]):
    """
    Process' result.
    """

    def __init__(self, rc: int, stdout: list[str], stderr: list[str], error: ProcessErrorType) -> None:
        """
        :param rc: Return code.
        :type rc: int
        :param stdout: Standard output, line by line.
        :type stdout: list[str]
        :param stderr: Standard error output, line by line.
        :type stderr: list[str]
        :param error: Process error object that can be raised manually with :meth:`throw`.
        :type error: ProcessErrorType
        """
        self.rc: int = rc
        """Return code."""

        self.stdout: str = "\n".join(stdout)
        """Standard output."""

        self.stderr: str = "\n".join(stderr)
        """Standard error output."""

        self.stdout_lines: list[str] = stdout
        """Standard output, line by line."""

        self.stderr_lines: list[str] = stderr
        """Standard error output, line by line."""

        self.error: ProcessErrorType = error
        """Process error object that can be raised manually with :meth:`throw`."""

    def throw(self) -> NoReturn:
        """
        Raise ProcessErrorType for this command manually.

        The error is available to raise even if return code is 0. Therefore the
        caller may choose to raise the error on any condition, not just the
        return code.

        :raises ValueError: If no error is set.
        :raises ProcessErrorType: Error generated for this process result.
        """
        if self.error is None:
            raise ValueError("No error is set.")

        raise self.error


ProcessResultType = TypeVar("ProcessResultType", bound=ProcessResult)
"""Generic process result type. Must be a subclass of :class:`ProcessResult`."""


ProcessInputBufferType = TypeVar("ProcessInputBufferType", bound=ProcessInputBuffer)
"""Generic process input buffer type. Must be a subclass of :class:`ProcessInputBuffer`."""


class Process(ABC, Generic[ProcessResultType, ProcessInputBufferType, ProcessTimeoutErrorType]):
    """
    Process manager.
    """

    _genid = itertools.count()

    def __init__(
        self,
        *,
        command: str,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | bytes | None = None,
        shell: Shell,
        logger: MultihostLogger,
        log_level: ProcessLogLevel,
        timeout: int,
        blocking_call: bool,
        additional_log_data: dict[str, Any] | None = None,
    ) -> None:
        """
        :param command: Command to execute.
        :type command: str
        :param cwd: Working directory, defaults to None
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | bytes | None, optional
        :param shell: Shell that will execute the command
        :type shell: Shell
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param log_level: Log level.
        :type log_level: ProcessLogLevel
        :param timeout: Timeout in seconds, value ``0`` means that timeout is
            disabled.
        :type timeout: int
        :param blocking_call: Is this a blocking execution?
        :type blocking_call: bool
        :param additional_log_data: Additional data that will be added to the
            log messages, defaults to None
        :type additional_log_data: dict[str, Any] | None, optional
        """
        self.id: int = next(self._genid) + 1
        """Autogenerated command ID."""

        self.command: str = textwrap.dedent(command).strip()
        """Executed command."""

        self.cwd: str | None = cwd
        """Working directory."""

        self.env: dict[str, Any] = env if env is not None else {}
        """Additional environment variables."""

        self.input: str | bytes | None = input
        """Input data."""

        self.shell: Shell = shell
        """Shell that executes the command."""

        self.logger: MultihostLogger = logger
        """Multihost logger."""

        self.log_level: ProcessLogLevel = log_level
        """Log level."""

        self.timeout: int = timeout
        """Timeout."""

        self.blocking_call: bool = blocking_call
        """True if this is a blocking call."""

        self.additional_log_data: dict[str, Any] = additional_log_data if additional_log_data is not None else {}
        """Additional data that will be added to the log messages."""

        self.full_command_line: str = self.shell.build_command_line(self.command, cwd=self.cwd, env=self.env)
        """Full command line that will be executed."""

        # Overwrite log level if requested.
        debug = os.getenv("MH_CONNECTION_DEBUG", "no")
        if debug.lower() in ["true", "yes", "1"]:
            self.log_level = ProcessLogLevel.Full

    @property
    @abstractmethod
    def in_progress(self) -> bool:
        """
        :return: True if the process is already started and running.
        :rtype: bool
        """
        pass

    @property
    @abstractmethod
    def stdout(self) -> Generator[str, None, None]:
        """
        Standard output, returns generator which yields output line by line.

        .. code-block:: python

            # Read single line, this will block until there is a line to read or
            # EOF is reached
            line = next(process.stdout)

            # Read all lines, this will block until EOF or EOF is reached
            lines = list(process.stdout)

            # Iterate over all lines
            for line in process.stdout:
                pass

        :raises RuntimeError: If the process is not running.
        :return: Standard output generator.
        :rtype: Generator[str, None, None]
        """
        pass

    @property
    @abstractmethod
    def stderr(self) -> Generator[str, None, None]:
        """
        Standard error output, returns generator which yields error output line
        by line.

        .. code-block:: python

            # Read single line, this will block until there is a line to read or
            # EOF is reached
            line = next(process.stderr)

            # Read all lines, this will block until EOF or EOF is reached
            lines = list(process.stderr)

            # Iterate over all lines
            for line in process.stderr:
                pass

        :raises RuntimeError: If the process is not running.
        :return: Standard error output generator.
        :rtype: Generator[str, None, None]
        """
        pass

    @property
    @abstractmethod
    def stdin(self) -> ProcessInputBufferType:
        """
        Command's standard input.

        Call :meth:`send_eof` to close the standard input and notify the process
        that there will be no more input data.

        .. code-block:: python

            # Write data
            process.stdin.write('Hello World')

            # Send EOF to indicate that there will be no more input data.
            process.send_eof()

        :raises RuntimeError: If the process is not running.
        :return: Standard input file.
        :rtype: ProcessInputBufferType
        """
        pass

    def run(self) -> Self:
        """
        Execute the command.

        :return: Self.
        :rtype: Self
        """
        if self.log_level in (ProcessLogLevel.Short, ProcessLogLevel.Full):
            self.logger.info(
                self.__msg_execution(),
                extra={
                    "data": {
                        **self.additional_log_data,
                        "Shell": self.shell.shell_command,
                        "Command": self.command,
                        "Input": self.input,
                        "Working directory": self.cwd,
                        "Extra environment": self.env,
                    }
                },
            )

        self._run()

        return self

    def wait(self, *, raise_on_error: bool = True, timeout: int | None = None) -> ProcessResultType:
        """
        Wait for the command to finish.

        EOF is send to standard input to indicate that there will be no
        additional input data. Then it waits for the command to finish.

        :param raise_on_error: If True, :class:`ProcessError` is raised on
            non-zero return code, defaults to True
        :type raise_on_error: bool, optional
        :param timeout: Timeout in seconds (default to ``None`` = use default
            value, ``0`` means no timeout)
        :type timeout: int | None
        :raises ProcessError: If ``raise_on_error`` is True and the command
            exited with non-zero return code.
        :raises ProcessTimeoutErrorType: If timeout occurred.
        :return: Command result.
        :rtype: ProcessResultType
        """
        if not self.in_progress:
            raise RuntimeError("Calling wait on process that has not yet started.")

        timeout = self.timeout if timeout is None else timeout

        try:
            result = timeout_decorator(timeout)(self._wait)()
        except ProcessTimeoutError as e:
            self._log_timeout(e)
            raise

        self._log_result(result)

        if raise_on_error and result.rc != 0:
            result.throw()

        return result

    def _log_result(self, result: ProcessResult) -> None:
        # Split the logging for a sync and async call.
        #
        # Synchronous/blocking calls have different message format to avoid
        # repeating information - we only want to print the result as the
        # command that was executed is logged in the previous line.
        #
        # Asynchronous/non-blocking calls produce full message with the command
        # that was executed and its result since the first message about the
        # command being executed is already lost in previous logs.
        #
        # We want to follow the async message for ProcessLogLevel.Error, since
        # this level does not produce any message about running the command. It
        # only logs the result if the command fails. Therefore, we want to
        # produce the full message.

        if self.blocking_call and self.log_level != ProcessLogLevel.Error:
            self._log_result_sync(result)
        else:
            self._log_result_async(result)

    def _log_result_sync(self, result: ProcessResult) -> None:
        msg = "Previous command completed with exit code " + self.__msg_rc(result.rc)

        match self.log_level:
            case ProcessLogLevel.Short:
                self.logger.info(msg)
            case ProcessLogLevel.Full:
                self.logger.info(
                    msg,
                    extra={
                        "data": {
                            "Output": result.stdout,
                            "Error output": result.stderr,
                        }
                    },
                )
            case _:
                pass

    def _log_result_async(self, result: ProcessResult) -> None:
        msg = (
            self.logger.colorize("Command ", c.Style.BRIGHT)
            + self.__msg_id()
            + self.logger.colorize(" completed with exit code ", c.Style.BRIGHT)
            + self.__msg_rc(result.rc)
        )

        match self.log_level:
            case ProcessLogLevel.Short:
                self.logger.info(
                    msg,
                    extra={
                        "data": {
                            **self.additional_log_data,
                            "Shell": self.shell.shell_command,
                            "Command": self.command,
                            "Input": self.input,
                            "Working directory": self.cwd,
                            "Extra environment": self.env,
                        }
                    },
                )
            case ProcessLogLevel.Full:
                self.logger.info(
                    msg,
                    extra={
                        "data": {
                            **self.additional_log_data,
                            "Shell": self.shell.shell_command,
                            "Command": self.command,
                            "Input": self.input,
                            "Working directory": self.cwd,
                            "Extra environment": self.env,
                            "Output": result.stdout,
                            "Error output": result.stderr,
                        }
                    },
                )
            case ProcessLogLevel.Error if result.rc != 0:
                self.logger.info(
                    msg,
                    extra={
                        "data": {
                            **self.additional_log_data,
                            "Shell": self.shell.shell_command,
                            "Command": self.command,
                            "Input": self.input,
                            "Working directory": self.cwd,
                            "Extra environment": self.env,
                            "Output": result.stdout,
                            "Error output": result.stderr,
                        }
                    },
                )
            case _:
                pass

    def _log_timeout(self, error: ProcessTimeoutError) -> None:
        msg = (
            self.logger.colorize("Command ", c.Style.BRIGHT)
            + self.__msg_id()
            + self.logger.colorize(f" timed out after {error.timeout} seconds", c.Style.BRIGHT)
        )

        self.logger.error(
            msg,
            extra={
                "data": {
                    **self.additional_log_data,
                    "Shell": self.shell.shell_command,
                    "Command": self.command,
                    "Input": self.input,
                    "Working directory": self.cwd,
                    "Extra environment": self.env,
                    "Output": error.stdout,
                    "Error output": error.stderr,
                }
            },
        )

    @abstractmethod
    def _run(self) -> None:
        """
        Execute the command.

        This is an internal method called by :meth:`run` after executing
        generic code.
        """
        pass

    @abstractmethod
    def _wait(self) -> ProcessResultType:
        """
        Wait for the command to finish.

        EOF is send to standard input to indicate that there will be no
        additional input data. Then it waits for the command to finish.

        This is an internal method called by :meth:`run` after executing
        generic code.

        :return: Command result.
        :rtype: ProcessResultType
        """
        pass

    @abstractmethod
    def send_eof(self) -> None:
        """
        Send EOF to standard input to indicate that there will be no more
        input data.

        :raises RuntimeError: If the process is not running.
        """
        pass

    @abstractmethod
    def send_signal(self, sig: signal.Signals) -> None:
        """
        Send signal to the running process.

        :param sig: A signal constant from ``signal``, e.g. ``signal.SIGUSR1``.
        :type sig: signal.Signals
        :raises RuntimeError: If the process is not running.
        """
        pass

    @abstractmethod
    def _close(self) -> None:
        """
        Close the communication channel.

        The communication channel is forcefully closed even if the process
        is still running.

        :return: Command result.
        :rtype: ProcessResultType
        """
        pass

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *exc_info):
        self.wait()

    def __msg_id(self) -> str:
        return self.logger.colorize(f"#{self.id}", c.Style.BRIGHT, c.Fore.BLUE)

    def __msg_rc(self, rc: int) -> str:
        if rc == 0:
            return self.logger.colorize(rc, c.Style.BRIGHT, c.Fore.GREEN)

        return self.logger.colorize(rc, c.Style.BRIGHT, c.Fore.RED)

    def __msg_execution(self) -> str:
        return f'{self.logger.colorize("Executing command", c.Style.BRIGHT)} ' + self.__msg_id()


ProcessType = TypeVar("ProcessType", bound=Process)
"""Generic process type. Must be a subclass of :class:`Process`."""


class ConnectionError(Exception):
    """
    Unable to connect to the host.
    """

    pass


class Connection(ABC, Generic[ProcessType, ProcessResultType]):
    """
    Abstract connection to the host.

    This is an abstract class. The following examples utilizes
    :class:`ssh.SSHClient` which implements this interface.

    .. code-block:: python
        :caption: Example: Blocking call

        # Connect to SSH server, it is automatically disconnected when leaving
        the with statement with SSHClient(host, user=username,
        password=password, logger=logger) as ssh:
            result = ssh.run('echo Hello World') print(result.rc)
            print(result.stdout)

            result = ssh.run('cat', input='Hello World') print(result.rc)
            print(result.stdout)

    .. code-block:: python
        :caption: Example: Non-blocking call

        # Connect to SSH server, it is automatically disconnected when leaving
        the with statement with SSHClient(host, user=username,
        password=password, logger=logger) as ssh:
            # The process is executed, but it does not block. In order to wait
            for it to finish, run process.wait() process = ssh.async_run('echo
            Hello World') result = process.wait() print(result.rc)
            print(result.stdout)

            # You can write to stdin directly in asynchronous run process =
            ssh.async_run('cat') process.stdin.write('Hello World')
            process.send_eof() result = process.wait() print(result.rc)
            print(result.stdout)

            # You can also work with inputs and outputs more interactively. #
            The process is automatically waited when leaving the with statement.
            with ssh.async_run('bash') as process:
                process.stdin.write('echo Hello World\\n')
                print(next(process.stdout))

                process.stdin.write('echo This works as well\\n')
                print(next(process.stdout))

    .. note::

        It is possible to set ``MH_CONNECTION_DEBUG=yes`` environment variable
        to log output and exit status to from commands, regardless of what log
        level is used. This essentially enforces the
        :attr:`ProcessLogLevel.Full` level.
    """

    def __init__(self, *, shell: Shell, logger: MultihostLogger, timeout: int = 300) -> None:
        """
        :param shell: Shell used to run commands and scripts.
        :type shell: str, optional
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param timeout: Timeout in seconds (defaults to 300), value
            ``0`` means that timeout is disabled.
        :type timeout: int

        """
        self.shell: Shell = shell
        """Shell used to run commands and scripts."""

        self.logger: MultihostLogger = logger
        """Multihost logger."""

        self.timeout: int = timeout
        """Default timeout for command execution."""

    @property
    @abstractmethod
    def connected(self) -> bool:
        """
        :return: True if the connection is established, False otherwise.
        :rtype: bool
        """
        pass

    @abstractmethod
    def connect(self) -> None:
        """
        Connect to the host.

        :raises ConnectionError: If connection can not be established.
        """
        pass

    @abstractmethod
    def disconnect(self) -> None:
        """
        Disconnect.
        """
        pass

    @abstractmethod
    def create_process(
        self,
        *,
        command: str,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | bytes | None = None,
        log_level: ProcessLogLevel,
        blocking_call: bool,
    ) -> ProcessType:
        """
        Create a new process.

        :param command: Command to execute.
        :type command: str
        :param cwd: Working directory, defaults to None
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | bytes | None, optional
        :param log_level: Log level.
        :type log_level: ProcessLogLevel
        :param blocking_call: Is this a blocking execution?
        :type blocking_call: bool
        :return: Newly created process that is not yet running.
        :rtype: ProcessType
        """
        pass

    def async_run(
        self,
        command: str,
        *,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | bytes | None = None,
        log_level: ProcessLogLevel = ProcessLogLevel.Full,
    ) -> ProcessType:
        """
        Non-blocking command call.

        The command is run under shell specified in the constructor and it is
        executed immediately, however it does not wait for the command to
        finish.

        :param command: Command to run.
        :type command: str
        :param cwd: Working directory, defaults to None (= do not change)
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | bytes | None, optional
        :param log_level: Log level, defaults to ProcessLogLevel.Full
        :type log_level: ProcessLogLevel, optional
        :return: Instance of :class:`Process`, the process is already running.
        :rtype: ProcessType
        """
        if not isinstance(command, str):
            raise ValueError("Parameter command is not a string, did you mean async_exec() instead of async_run()?")

        self.connect()

        process = self.create_process(
            command=command,
            cwd=cwd,
            env=env,
            input=input,
            log_level=log_level,
            blocking_call=False,
        )

        process.run()
        return process

    def run(
        self,
        command: str,
        *,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | None = None,
        log_level: ProcessLogLevel = ProcessLogLevel.Full,
        raise_on_error: bool = True,
        timeout: int | None = None,
    ) -> ProcessResultType:
        """
        Blocking command call.

        The command is run under shell specified in the constructor and it is
        executed immediately. It waits for the command to finish and returns its
        result.

        :param command: Command to run.
        :type command: str
        :param cwd: Working directory, defaults to None (= do not change)
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | None, optional
        :param log_level: Log level, defaults to ProcessLogLevel.Full
        :type log_level: ProcessLogLevel, optional
        :param raise_on_error: If True, raise :class:`ProcessError` if command
            exited with non-zero return code, defaults to True
        :type raise_on_error: bool, optional
        :param timeout: Timeout in seconds (defaults to ``None`` = use default
            connection timeout)
        :type timeout: int | None
        :raises ProcessError: If ``raise_on_error`` is True and the command
            exited with non-zero return code.
        :return: Command result.
        :rtype: ProcessResultType
        """
        if not isinstance(command, str):
            raise ValueError("Parameter command is not a string, did you mean exec() instead of run()?")

        self.connect()

        process = self.create_process(
            command=command,
            cwd=cwd,
            env=env,
            input=input,
            log_level=log_level,
            blocking_call=True,
        )

        process.run()

        return process.wait(raise_on_error=raise_on_error, timeout=timeout)

    def async_exec(
        self,
        argv: list[Any],
        *,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | None = None,
        log_level: ProcessLogLevel = ProcessLogLevel.Full,
    ) -> ProcessType:
        """
        Non-blocking command call.

        The command is run under shell specified in the constructor and it is
        executed immediately, however it does not wait for the command to
        finish.

        The command is provided as ``argv`` list.

        :param argv: Command to run.
        :type argv: list[Any]
        :param cwd: Working directory, defaults to None (= do not change)
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | None, optional
        :param log_level: Log level, defaults to ProcessLogLevel.Full
        :type log_level: ProcessLogLevel, optional
        :return: Instance of :class:`Process`, the process is already running.
        :rtype: ProcessType
        """
        if not isinstance(argv, list):
            raise ValueError("Parameter argv is not a list, did you mean async_run() instead of async_exec()?")

        argv = [str(x) for x in argv]
        command = shlex.join(argv)

        return self.async_run(
            command,
            cwd=cwd,
            env=env,
            input=input,
            log_level=log_level,
        )

    def exec(
        self,
        argv: list[Any],
        *,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | None = None,
        log_level: ProcessLogLevel = ProcessLogLevel.Full,
        raise_on_error: bool = True,
        timeout: int | None = None,
    ) -> ProcessResultType:
        """
        Blocking command call.

        The command is run under shell specified in the constructor and it is
        executed immediately. It waits for the command to finish and returns its
        result.

        The command is provided as ``argv`` list.

        :param argv: Command to run.
        :type argv: list[Any]
        :param cwd: Working directory, defaults to None (= do not change)
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | None, optional
        :param log_level: Log level, defaults to ProcessLogLevel.Full
        :type log_level: ProcessLogLevel, optional
        :param raise_on_error: If True, raise :class:`ProcessError` if command
            exited with non-zero return code, defaults to True
        :type raise_on_error: bool, optional
        :param timeout: Timeout in seconds (defaults to ``None`` = use default
            connection timeout)
        :type timeout: int | None
        :raises ProcessError: If ``raise_on_error`` is True and the command
            exited with non-zero return code.
        :return: Command result.
        :rtype: ProcessResultType
        """
        if not isinstance(argv, list):
            raise ValueError("Parameter argv is not a list, did you mean run() instead of exec()?")

        argv = [str(x) for x in argv]
        command = shlex.join(argv)

        return self.run(
            command,
            cwd=cwd,
            env=env,
            input=input,
            log_level=log_level,
            raise_on_error=raise_on_error,
            timeout=timeout,
        )

    def expect(
        self,
        expect_script: str,
        *,
        verbose: bool = True,
        raise_on_error: bool = False,
    ) -> ProcessResultType:
        """
        Run expect script.

        :param expect_script: Expect script.
        :type expect_script: str
        :param verbose: Enable expect debug output (-d), default to True.
        :type verbose: bool, optional
        :param raise_on_error: If True, raise :class:`ProcessError` if command
            exited with non-zero return code, defaults to False
        :type raise_on_error: bool, optional
        :return: Expect script result.
        :rtype: ProcessResultType
        """
        args = ["-d"] if verbose else []
        return self.exec(["/bin/expect", *args], input=expect_script, raise_on_error=raise_on_error)

    def expect_nobody(
        self,
        expect_script: str,
        *,
        verbose: bool = True,
        raise_on_error: bool = False,
    ) -> ProcessResultType:
        """
        Run expect script as user nobody.

        The main use case is to avoid running the command as root if the client
        is connected to the root user SSH session.

        :param expect_script: Expect script.
        :type expect_script: str
        :param verbose: Enable expect debug output (-d), default to True.
        :type verbose: bool, optional
        :param raise_on_error: If True, raise :class:`ProcessError` if command
            exited with non-zero return code, defaults to False
        :type raise_on_error: bool, optional
        :return: Expect return code.
        :rtype: ProcessResultType
        """
        args = " -d" if verbose else ""
        return self.run(
            f'su --shell /bin/sh nobody -c "/bin/expect{args}"', input=expect_script, raise_on_error=raise_on_error
        )

    def __enter__(self) -> Connection:
        """
        Connect to the host.

        :return: Connection instance.
        :rtype: Connection
        """
        self.connect()
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """
        Disconnect.
        """
        self.disconnect()

    @classmethod
    @abstractmethod
    def from_confdict(cls, host: MultihostHost, confdict: dict[str, Any]) -> Self:
        """
        Create new instance of this class from configuration dictionary.

        :param host: Multihost host that will use this connection.
        :type host: MultihostHost
        :param confdict: Configuration dictionary.
        :type confdict: dict[str, Any]
        :return: New instance.
        :rtype: Self
        """
        pass


class Shell(ABC):
    """
    Multihost shell abstraction.
    """

    def __init__(self, name: str, shell_command: str) -> None:
        """
        :param name: Shell name.
        :type name: str
        :param shell_command: Shell command that will execute user scripts.
        :param shell_command: str
        """
        self.name: str = name
        """Shell name."""

        self.shell_command: str = shell_command
        """Shell command that will execute user scripts."""

    @abstractmethod
    def build_command_line(self, script: str, *, cwd: str | None, env: dict[str, Any]) -> str:
        """
        Create a complete command that will execute given script in this shell.

        User is able to specify a different working directory and additional
        environment variables. This method adds this information to the user's
        command and returns a full script that will switch the working
        directory, sets environment variables and then executes the given
        command or script.

        :param script: Script
        :type script: str
        :param cwd: Working directory, ``None`` means no change.
        :type cwd: str | None
        :param env: Additional environment variables.
        :type env: dict[str, Any]
        :return: Complete command line to run to execute the script in this
            shell.
        :rtype: str
        """
        pass


class Bash(Shell):
    """
    Bash shell abstraction.
    """

    def __init__(self) -> None:
        super().__init__("bash", "/usr/bin/env bash -c")

    def build_command_line(self, script: str, *, cwd: str | None, env: dict[str, Any]) -> str:
        full_script = self._add_cwd_and_env(script, cwd=cwd, env=env)
        escaped_script = self._escape_single_quotes(full_script)

        return f"{self.shell_command} '{escaped_script}'"

    def _add_cwd_and_env(self, script: str, *, cwd: str | None, env: dict[str, Any]) -> str:
        out = ""

        # Set environment variables
        for key, value in env.items():
            out += f"export {key}={shlex.quote(str(value))}\n"

        # Set working directory
        if cwd is not None:
            out += f"cd {shlex.quote(cwd)}\n"

        if out:
            out += "\n"

        out += script

        return out

    def _escape_single_quotes(self, command: str) -> str:
        """
        We call the command as `/usr/bin/env bash -c '$command'`.

        We need to escape ' inside the script to make it work correctly.
        """
        return command.replace("'", "'\"'\"'")


class Powershell(Shell):
    """
    Powershell shell abstraction.
    """

    def __init__(self) -> None:
        super().__init__("powershell", "powershell -NonInteractive -Command")

    def build_command_line(self, script: str, *, cwd: str | None, env: dict[str, Any]) -> str:
        full_script = self._add_cwd_and_env(script, cwd=cwd, env=env)
        escaped_script = self._escape_quotes(full_script)

        return f"{self.shell_command} '{escaped_script}'"

    def _add_cwd_and_env(self, script: str, *, cwd: str | None, env: dict[str, Any]) -> str:
        out = ""

        # Set environment variables
        for key, value in env.items():
            out += f"$Env:{key} = {shlex.quote(str(value))}\n"

        # Set working directory
        if cwd is not None:
            out += f"cd {shlex.quote(cwd)}\n"

        if out:
            out += "\n"

        out += script

        return out

    def _escape_quotes(self, command: str) -> str:
        """
        We need to escape quotes inside the script to make it work correctly.
        """
        return command.replace("'", "''").replace('"', '\\"')
